"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var standard_1 = require("./standard");
var typeguards_1 = require("./typeguards");
exports.UniversalVarType = "struct js_var";
exports.VoidType = "void";
exports.PointerVarType = "void *";
exports.StringVarType = "const char *";
exports.NumberVarType = "int16_t";
exports.BooleanVarType = "uint8_t";
exports.RegexVarType = "struct regex_struct_t";
exports.RegexMatchVarType = "struct regex_match_struct_t";
var getTypeBodyText = function (t) { return typeof t === "string" ? t : t.getBodyText(); };
/** Type that represents static or dynamic array */
var ArrayType = /** @class */ (function () {
    function ArrayType(elementType, capacity, isDynamicArray) {
        this.elementType = elementType;
        this.capacity = capacity;
        this.isDynamicArray = isDynamicArray;
    }
    ArrayType.getArrayStructName = function (elementTypeText) {
        while (elementTypeText.indexOf(exports.NumberVarType) > -1)
            elementTypeText = elementTypeText.replace(exports.NumberVarType, "number");
        while (elementTypeText.indexOf(exports.StringVarType) > -1)
            elementTypeText = elementTypeText.replace(exports.StringVarType, "string");
        while (elementTypeText.indexOf(exports.PointerVarType) > -1)
            elementTypeText = elementTypeText.replace(exports.PointerVarType, "pointer");
        while (elementTypeText.indexOf(exports.BooleanVarType) > -1)
            elementTypeText = elementTypeText.replace(exports.BooleanVarType, "bool");
        elementTypeText = elementTypeText.replace(/^struct ([a-z0-9_]+)_t \*$/, function (all, g1) { return g1; }).replace(/^struct js_var/, "js_var");
        return "array_" +
            elementTypeText
                .replace(/^static /, '').replace('{var}', '').replace(/[\[\]]/g, '')
                .replace(/ /g, '_')
                .replace(/const char \*/g, 'string')
                .replace(/\*/g, 'p') + "_t";
    };
    ArrayType.prototype.getText = function () {
        var elementType = this.elementType;
        var elementTypeText;
        if (typeof elementType === 'string')
            elementTypeText = elementType;
        else
            elementTypeText = elementType.getText();
        var structName = ArrayType.getArrayStructName(elementTypeText);
        if (this.isDynamicArray)
            return "struct " + structName + " *";
        else if (elementTypeText.indexOf('{var}') > -1)
            return elementTypeText + "[" + this.capacity + "]";
        else
            return "static " + elementTypeText + " {var}[" + this.capacity + "]";
    };
    ArrayType.prototype.getBodyText = function () {
        return getTypeBodyText(this.elementType) + "[" + (this.isDynamicArray ? "" : this.capacity) + "]";
    };
    return ArrayType;
}());
exports.ArrayType = ArrayType;
/** Type that represents JS object with static properties (implemented as C struct) */
var StructType = /** @class */ (function () {
    function StructType(propertyDefs) {
        this.propertyDefs = propertyDefs;
    }
    StructType.prototype.getText = function () {
        return this.forcedType || 'struct ' + this.structName + ' *';
    };
    Object.defineProperty(StructType.prototype, "properties", {
        get: function () {
            var _this = this;
            return Object.keys(this.propertyDefs)
                .sort(function (a, b) { return _this.propertyDefs[a].order - _this.propertyDefs[b].order; })
                .reduce(function (acc, k) { acc[k] = _this.propertyDefs[k].type; return acc; }, {});
        },
        enumerable: true,
        configurable: true
    });
    StructType.prototype.getBodyText = function () {
        var _this = this;
        return "{" + Object.keys(this.propertyDefs).sort().map(function (k) { return k + ": " + getTypeBodyText(_this.properties[k]); }).join("; ") + "}";
    };
    return StructType;
}());
exports.StructType = StructType;
/** Type that represents JS object with dynamic properties (implemented as dynamic dictionary) */
var DictType = /** @class */ (function () {
    function DictType(elementType) {
        this.elementType = elementType;
    }
    DictType.prototype.getText = function () {
        if (this.elementType == exports.UniversalVarType)
            return "struct dict_js_var_t *";
        else
            return "DICT(" + (typeof this.elementType === "string" ? this.elementType : this.elementType.getText()) + ")";
    };
    DictType.prototype.getBodyText = function () {
        return "{" + getTypeBodyText(this.elementType) + "}";
    };
    return DictType;
}());
exports.DictType = DictType;
var FuncType = /** @class */ (function () {
    function FuncType(returnType, parameterTypes, instanceType) {
        if (parameterTypes === void 0) { parameterTypes = []; }
        if (instanceType === void 0) { instanceType = null; }
        this.returnType = returnType;
        this.parameterTypes = parameterTypes;
        this.instanceType = instanceType;
    }
    FuncType.getReturnType = function (typeHelper, node) {
        var decl = typeHelper.getDeclaration(node);
        var type = typeHelper.getCType(decl);
        return type && type instanceof FuncType ? type.returnType : null;
    };
    FuncType.getInstanceType = function (typeHelper, node) {
        var decl = typeHelper.getDeclaration(node);
        var type = typeHelper.getCType(decl);
        return type && type instanceof FuncType ? type.instanceType : null;
    };
    FuncType.prototype.getText = function () {
        return typeof (this.returnType) === "string" ? this.returnType : this.returnType.getText();
    };
    FuncType.prototype.getBodyText = function () {
        var paramTypes = [].concat(this.parameterTypes);
        if (this.instanceType)
            paramTypes.unshift(this.instanceType);
        return getTypeBodyText(this.returnType) + "(" + paramTypes.map(function (pt) { return pt ? getTypeBodyText(pt) : exports.PointerVarType; }).join(", ") + ")";
    };
    return FuncType;
}());
exports.FuncType = FuncType;
exports.equalityOps = [
    ts.SyntaxKind.EqualsEqualsToken, ts.SyntaxKind.EqualsEqualsEqualsToken,
    ts.SyntaxKind.ExclamationEqualsToken, ts.SyntaxKind.ExclamationEqualsEqualsToken,
];
var relationalOps = [
    ts.SyntaxKind.GreaterThanToken, ts.SyntaxKind.GreaterThanEqualsToken,
    ts.SyntaxKind.LessThanToken, ts.SyntaxKind.LessThanEqualsToken
];
var logicalOps = [
    ts.SyntaxKind.BarBarToken, ts.SyntaxKind.AmpersandAmpersandToken
];
function operandsToNumber(leftType, op, rightType) {
    return typeguards_1.isNumberOp(op) || typeguards_1.isIntegerOp(op)
        || op == ts.SyntaxKind.PlusToken && !toNumberCanBeNaN(leftType) && !toNumberCanBeNaN(rightType)
        || exports.equalityOps.concat(relationalOps).indexOf(op) > -1 && (leftType !== exports.StringVarType || rightType !== exports.StringVarType);
}
exports.operandsToNumber = operandsToNumber;
function getBinExprResultType(leftType, op, rightType) {
    if (logicalOps.indexOf(op) > -1 || op === ts.SyntaxKind.EqualsToken)
        return rightType;
    if (relationalOps.indexOf(op) > -1 || exports.equalityOps.indexOf(op) > -1)
        return exports.BooleanVarType;
    if (leftType == null || rightType == null)
        return null;
    if (typeguards_1.isNumberOp(op) || typeguards_1.isIntegerOp(op))
        return toNumberCanBeNaN(leftType) || toNumberCanBeNaN(rightType) ? exports.UniversalVarType : exports.NumberVarType;
    if (op === ts.SyntaxKind.PlusToken || op === ts.SyntaxKind.PlusEqualsToken)
        return leftType === exports.UniversalVarType || rightType === exports.UniversalVarType ? exports.UniversalVarType
            : toPrimitive(leftType) === exports.StringVarType || toPrimitive(rightType) === exports.StringVarType ? exports.StringVarType
                : toPrimitive(leftType) === exports.NumberVarType && toPrimitive(rightType) == exports.NumberVarType ? exports.NumberVarType
                    : null;
    console.log("WARNING: unexpected binary expression!");
    return null;
}
exports.getBinExprResultType = getBinExprResultType;
function toNumberCanBeNaN(t) {
    return t !== null && t !== exports.PointerVarType && t !== exports.NumberVarType && t !== exports.BooleanVarType && !(t instanceof ArrayType && !t.isDynamicArray && t.capacity == 1 && !toNumberCanBeNaN(t.elementType));
}
exports.toNumberCanBeNaN = toNumberCanBeNaN;
function toPrimitive(t) {
    return t === null || t === exports.PointerVarType ? t : t === exports.NumberVarType || t === exports.BooleanVarType ? exports.NumberVarType : exports.StringVarType;
}
exports.toPrimitive = toPrimitive;
function findParentFunction(node) {
    var parentFunc = node;
    while (parentFunc && !ts.isFunctionDeclaration(parentFunc))
        parentFunc = parentFunc.parent;
    return parentFunc;
}
exports.findParentFunction = findParentFunction;
function findParentSourceFile(node) {
    var parent = node;
    while (!ts.isSourceFile(parent))
        parent = parent.parent;
    return parent;
}
exports.findParentSourceFile = findParentSourceFile;
var TypeHelper = /** @class */ (function () {
    function TypeHelper(typeChecker, allNodes) {
        this.typeChecker = typeChecker;
        this.allNodes = allNodes;
        this.arrayLiteralsTypes = {};
        this.objectLiteralsTypes = {};
        this.typeOfNodeDict = {};
        this.typesDict = {};
    }
    /** Get C type of TypeScript node */
    TypeHelper.prototype.getCType = function (node) {
        if (!node || !node.kind)
            return null;
        var found = this.typeOfNodeDict[node.pos + "_" + node.end];
        if (found)
            return found.type;
        switch (node.kind) {
            case ts.SyntaxKind.NumericLiteral:
                return exports.NumberVarType;
            case ts.SyntaxKind.TrueKeyword:
            case ts.SyntaxKind.FalseKeyword:
                return exports.BooleanVarType;
            case ts.SyntaxKind.StringLiteral:
                return exports.StringVarType;
            case ts.SyntaxKind.RegularExpressionLiteral:
                return exports.RegexVarType;
            case ts.SyntaxKind.ArrayLiteralExpression:
                {
                    if (!this.arrayLiteralsTypes[node.pos])
                        this.determineArrayType(node);
                    return this.arrayLiteralsTypes[node.pos];
                }
            case ts.SyntaxKind.ObjectLiteralExpression:
                {
                    if (!this.objectLiteralsTypes[node.pos])
                        this.objectLiteralsTypes[node.pos] = this.generateStructure(this.typeChecker.getTypeAtLocation(node));
                    return this.objectLiteralsTypes[node.pos];
                }
            case ts.SyntaxKind.CallExpression:
                {
                    var call = node;
                    var retType = standard_1.StandardCallHelper.getReturnType(this, call);
                    if (retType)
                        return retType;
                }
        }
        if (node.kind != ts.SyntaxKind.ImportClause && node.pos != -1) {
            var tsType = this.typeChecker.getTypeAtLocation(node);
            var type = tsType && this.convertType(tsType, node);
            if (type)
                return type;
        }
        return null;
    };
    /** Get textual representation of type of the parameter for inserting into the C code */
    TypeHelper.prototype.getTypeString = function (source) {
        var cType = source;
        if (source && source.flags != null && source.intrinsicName != null) // ts.Type
            cType = this.convertType(source);
        else if (source && source.flags != null && source.callSignatures != null && source.constructSignatures != null) // ts.Type
            cType = this.convertType(source);
        else if (source && source.kind != null && source.flags != null) // ts.Node
            cType = this.getCType(source);
        if (cType instanceof ArrayType) {
            return cType.getText();
        }
        else if (cType instanceof StructType)
            return cType.getText();
        else if (cType instanceof DictType)
            return cType.getText();
        else if (typeof cType === 'string')
            return cType;
        else
            return "/* Cannot determine variable type from source " + (source && source.getText ? source.getText() : JSON.stringify(source)) + "*/";
    };
    /** Postprocess TypeScript AST for better type inference and map TS types to C types */
    /** Creates typeOfNodeDict that is later used in getCType */
    TypeHelper.prototype.inferTypes = function () {
        var _this = this;
        var type = function (t) { return ({ getType: typeof (t) === "string" ? function (_) { return t; } : t }); };
        var struct = function (prop, pos, elemType) {
            if (elemType === void 0) { elemType = exports.PointerVarType; }
            return new StructType((_a = {}, _a[prop] = { type: elemType, order: pos }, _a));
            var _a;
        };
        var typeEqualities = [];
        var addEquality = function (typeGuard, node1, node2) {
            if (typeof node2 == "function")
                typeEqualities.push([typeGuard, node1, { getNode: node2 }]);
            else
                typeEqualities.push([typeGuard, node1, node2]);
        };
        // expressions
        addEquality(ts.isIdentifier, function (n) { return n; }, function (n) { return _this.getDeclaration(n); });
        addEquality(typeguards_1.isEqualsExpression, function (n) { return n.left; }, function (n) { return n.right; });
        addEquality(ts.isConditionalExpression, function (n) { return n.whenTrue; }, function (n) { return n.whenFalse; });
        addEquality(ts.isConditionalExpression, function (n) { return n; }, function (n) { return n.whenTrue; });
        addEquality(typeguards_1.isConvertToNumberExpression, function (n) { return n; }, type(function (n) { return _this.getCType(n.operand) == exports.NumberVarType ? exports.NumberVarType : exports.UniversalVarType; }));
        addEquality(ts.isBinaryExpression, function (n) { return n; }, type(function (n) { return getBinExprResultType(_this.getCType(n.left), n.operatorToken.kind, _this.getCType(n.right)); }));
        addEquality(ts.isBinaryExpression, function (n) { return n.left; }, type(function (n) {
            var resultType = _this.getCType(n);
            var operandType = _this.getCType(n.left);
            if (resultType === exports.UniversalVarType) {
                return typeguards_1.isCompoundAssignment(n.operatorToken) ? exports.UniversalVarType
                    : operandType instanceof ArrayType ? new ArrayType(exports.UniversalVarType, 0, true)
                        : operandType instanceof StructType || operandType instanceof DictType ? new DictType(exports.UniversalVarType)
                            : null;
            }
            else
                return null;
        }));
        addEquality(ts.isBinaryExpression, function (n) { return n.right; }, type(function (n) {
            var resultType = _this.getCType(n);
            var operandType = _this.getCType(n.right);
            if (resultType === exports.UniversalVarType) {
                return operandType instanceof ArrayType ? new ArrayType(exports.UniversalVarType, 0, true)
                    : operandType instanceof StructType || operandType instanceof DictType ? new DictType(exports.UniversalVarType)
                        : null;
            }
            else
                return null;
        }));
        addEquality(typeguards_1.isNullOrUndefinedOrNaN, function (n) { return n; }, type(exports.UniversalVarType));
        addEquality(ts.isParenthesizedExpression, function (n) { return n; }, function (n) { return n.expression; });
        addEquality(ts.isVoidExpression, function (n) { return n; }, type(exports.UniversalVarType));
        addEquality(ts.isVoidExpression, function (n) { return n.expression; }, type(exports.PointerVarType));
        addEquality(ts.isTypeOfExpression, function (n) { return n; }, type(exports.StringVarType));
        addEquality(typeguards_1.isDeleteExpression, function (n) { return n; }, type(exports.BooleanVarType));
        addEquality(typeguards_1.isDeleteExpression, function (n) { return n.expression.expression; }, type(function (n) { return new DictType(exports.UniversalVarType); }));
        // fields
        addEquality(ts.isPropertyAssignment, function (n) { return n; }, function (n) { return n.initializer; });
        addEquality(ts.isPropertyAssignment, function (n) { return n.parent; }, type(function (n) {
            var propName = (ts.isIdentifier(n.name) || ts.isStringLiteral(n.name)) && n.name.text;
            return struct(propName, n.pos, _this.getCType(n) || exports.PointerVarType);
        }));
        addEquality(ts.isPropertyAssignment, function (n) { return n; }, type(function (n) {
            var propName = (ts.isIdentifier(n.name) || ts.isStringLiteral(n.name)) && n.name.text;
            var type = _this.getCType(n.parent);
            return type instanceof StructType ? type.properties[propName] : null;
        }));
        addEquality(typeguards_1.isFieldPropertyAccess, function (n) { return n.expression; }, type(function (n) { return struct(n.name.getText(), n.pos, _this.getCType(n) || exports.PointerVarType); }));
        addEquality(typeguards_1.isFieldPropertyAccess, function (n) { return n; }, type(function (n) {
            var type = _this.getCType(n.expression);
            return type instanceof StructType ? type.properties[n.name.getText()]
                : type instanceof ArrayType && n.name.getText() == "length" ? exports.NumberVarType
                    : type === exports.StringVarType && n.name.getText() == "length" ? exports.NumberVarType
                        : type instanceof ArrayType || type instanceof DictType ? type.elementType
                            : type === exports.UniversalVarType && n.name.getText() == "length" ? exports.NumberVarType
                                : type === exports.UniversalVarType ? exports.UniversalVarType
                                    : null;
        }));
        addEquality(typeguards_1.isFieldElementAccess, function (n) { return n.expression; }, type(function (n) {
            var type = _this.getCType(n.argumentExpression);
            var elementType = _this.getCType(n) || exports.PointerVarType;
            return ts.isStringLiteral(n.argumentExpression) ? struct(n.argumentExpression.getText().slice(1, -1), n.pos, elementType)
                : ts.isNumericLiteral(n.argumentExpression) ? new ArrayType(elementType, 0, false)
                    : type == exports.NumberVarType ? new ArrayType(elementType, 0, false)
                        : type == exports.StringVarType ? new DictType(elementType)
                            : null;
        }));
        addEquality(typeguards_1.isFieldElementAccess, function (n) { return n; }, type(function (n) {
            var type = _this.getCType(n.expression);
            return ts.isStringLiteral(n.argumentExpression) && type instanceof StructType ? type.properties[n.argumentExpression.getText().slice(1, -1)]
                : ts.isStringLiteral(n.argumentExpression) && type instanceof ArrayType && n.argumentExpression.getText().slice(1, -1) == "length" ? exports.NumberVarType
                    : ts.isStringLiteral(n.argumentExpression) && type === exports.StringVarType && n.argumentExpression.getText().slice(1, -1) == "length" ? exports.NumberVarType
                        : ts.isStringLiteral(n.argumentExpression) && type === exports.UniversalVarType && n.argumentExpression.getText().slice(1, -1) == "length" ? exports.NumberVarType
                            : type instanceof ArrayType || type instanceof DictType ? type.elementType
                                : type === exports.UniversalVarType ? exports.UniversalVarType
                                    : null;
        }));
        var _loop_1 = function (i_1) {
            addEquality(ts.isArrayLiteralExpression, function (n) { return n; }, type(function (n) {
                var elemType = _this.getCType(n.elements[i_1]);
                return elemType ? new ArrayType(elemType, 0, false) : null;
            }));
            addEquality(ts.isArrayLiteralExpression, function (n) { return n.elements[i_1]; }, type(function (n) {
                var arrType = _this.getCType(n);
                return arrType && arrType instanceof ArrayType ? arrType.elementType
                    : arrType === exports.UniversalVarType ? exports.UniversalVarType
                        : null;
            }));
        };
        for (var i_1 = 0; i_1 < 10; i_1++) {
            _loop_1(i_1);
        }
        // calls
        addEquality(ts.isCallExpression, function (n) { return n.expression; }, function (n) { return _this.getDeclaration(n); });
        addEquality(ts.isCallExpression, function (n) { return n; }, type(function (n) { return FuncType.getReturnType(_this, n.expression); }));
        addEquality(ts.isCallExpression, function (n) { return n.expression; }, type(function (n) { return _this.getCType(n) ? new FuncType(_this.getCType(n)) : null; }));
        var _loop_2 = function (i_2) {
            addEquality(ts.isCallExpression, function (n) {
                var func = _this.getDeclaration(n.expression);
                return func ? func.parameters[i_2] : null;
            }, type(function (n) { return _this.getCType(n.arguments[i_2]); }));
        };
        for (var i_2 = 0; i_2 < 10; i_2++) {
            _loop_2(i_2);
        }
        addEquality(ts.isParameter, function (n) { return n; }, function (n) { return n.name; });
        addEquality(ts.isParameter, function (n) { return n; }, function (n) { return n.initializer; });
        addEquality(ts.isNewExpression, function (n) { return n; }, type(function (n) { return FuncType.getInstanceType(_this, n.expression); }));
        var _loop_3 = function (i_3) {
            addEquality(ts.isNewExpression, function (n) { return n.arguments[i_3]; }, function (n) {
                var func = _this.getDeclaration(n.expression);
                return func ? func.parameters[i_3] : null;
            });
        };
        for (var i_3 = 0; i_3 < 10; i_3++) {
            _loop_3(i_3);
        }
        addEquality(typeguards_1.isThisKeyword, function (n) { return findParentFunction(n); }, type(function (n) { return new FuncType(exports.VoidType, [], _this.getCType(n)); }));
        addEquality(typeguards_1.isThisKeyword, function (n) { return n; }, type(function (n) { return FuncType.getInstanceType(_this, findParentFunction(n)); }));
        addEquality(typeguards_1.isMethodCall, function (n) { return n.expression.expression; }, type(function (n) { return standard_1.StandardCallHelper.getObjectType(_this, n); }));
        addEquality(ts.isCallExpression, function (n) { return n; }, type(function (n) { return standard_1.StandardCallHelper.getReturnType(_this, n); }));
        var _loop_4 = function (i_4) {
            addEquality(ts.isCallExpression, function (n) { return n.arguments[i_4]; }, type(function (n) { return typeguards_1.isLiteral(n.arguments[i_4]) ? null : standard_1.StandardCallHelper.getArgumentTypes(_this, n)[i_4]; }));
        };
        for (var i_4 = 0; i_4 < 10; i_4++) {
            _loop_4(i_4);
        }
        // crutch for callback argument type in foreach
        addEquality(typeguards_1.isFunctionArgInMethodCall, function (n) { return n.parameters[0]; }, type(function (n) {
            var objType = _this.getCType(n.parent.expression.expression);
            return objType instanceof ArrayType && n.parent.expression.name.text == "forEach" ? objType.elementType : null;
        }));
        // statements
        addEquality(ts.isVariableDeclaration, function (n) { return n; }, function (n) { return n.initializer; });
        addEquality(ts.isFunctionDeclaration, function (n) { return n; }, type(function (n) { return new FuncType(exports.VoidType, n.parameters.map(function (p) { return _this.getCType(p); })); }));
        addEquality(typeguards_1.isForOfWithSimpleInitializer, function (n) { return n.expression; }, type(function (n) { return new ArrayType(_this.getCType(n.initializer.declarations[0]) || exports.PointerVarType, 0, false); }));
        addEquality(typeguards_1.isForOfWithSimpleInitializer, function (n) { return n.initializer.declarations[0]; }, type(function (n) {
            var type = _this.getCType(n.expression);
            return type instanceof ArrayType ? type.elementType : null;
        }));
        addEquality(typeguards_1.isForOfWithIdentifierInitializer, function (n) { return n.expression; }, type(function (n) { return new ArrayType(_this.getCType(n.initializer) || exports.PointerVarType, 0, false); }));
        addEquality(typeguards_1.isForOfWithIdentifierInitializer, function (n) { return n.initializer; }, type(function (n) {
            var type = _this.getCType(n.expression);
            return type instanceof ArrayType ? type.elementType : null;
        }));
        addEquality(ts.isForInStatement, function (n) { return n.initializer; }, type(exports.StringVarType));
        addEquality(ts.isReturnStatement, function (n) { return n.expression; }, type(function (n) { return FuncType.getReturnType(_this, findParentFunction(n)); }));
        addEquality(ts.isReturnStatement, function (n) { return findParentFunction(n); }, type(function (n) { return _this.getCType(n.expression) ? new FuncType(_this.getCType(n.expression)) : null; }));
        addEquality(ts.isCaseClause, function (n) { return n.expression; }, function (n) { return n.parent.parent.expression; });
        addEquality(ts.isCatchClause, function (n) { return n.variableDeclaration; }, type(exports.StringVarType));
        this.resolveTypes(typeEqualities);
    };
    TypeHelper.prototype.resolveTypes = function (typeEqualities) {
        var _this = this;
        this.allNodes.forEach(function (n) { return _this.setNodeType(n, _this.getCType(n)); });
        var equalities = [];
        typeEqualities.forEach(function (teq) {
            return _this.allNodes.forEach(function (node) { if (teq[0].bind(_this)(node))
                equalities.push([node, teq]); });
        });
        var changed;
        do {
            changed = false;
            for (var _i = 0, equalities_1 = equalities; _i < equalities_1.length; _i++) {
                var equality = equalities_1[_i];
                var node = equality[0], _a = equality[1], _ = _a[0], node1_func = _a[1], node2_resolver = _a[2];
                var node1 = node1_func(node);
                if (!node1)
                    continue;
                var type1 = this.getCType(node1);
                var node2 = node2_resolver.getNode ? node2_resolver.getNode(node) : null;
                var type2 = node2_resolver.getType ? node2_resolver.getType(node) : this.getCType(node2);
                if (!node2 && !type2)
                    continue;
                var _b = this.mergeTypes(type1, type2), type = _b.type, replaced = _b.replaced;
                if (type && replaced) {
                    if (type != type1)
                        changed = true;
                    if (node2 && type != type2)
                        changed = true;
                    this.setNodeType(node1, type);
                    if (node2)
                        this.setNodeType(node2, type);
                }
            }
        } while (changed);
        for (var k in this.typeOfNodeDict) {
            var type = this.typeOfNodeDict[k].type;
            if (type instanceof ArrayType && !type.isDynamicArray && type.capacity == 0)
                type.isDynamicArray = true;
        }
        /*
        this.allNodes
            .filter(n => !ts.isToken(n) && !ts.isBlock(n) && n.kind != ts.SyntaxKind.SyntaxList)
            .forEach(n => console.log(n.getText(), "|", ts.SyntaxKind[n.kind], "|", JSON.stringify(this.getCType(n))));
        
        this.allNodes
            .filter(n => ts.isIdentifier(n) && n.getText() == "string1")
            .forEach(n => console.log(
                n.getText(),
                "(" + n.parent.getText() + "/" + ts.SyntaxKind[n.parent.kind] + ")",
                "decl.", getDeclaration(this.typeChecker, n).getText() + "/" + ts.SyntaxKind[getDeclaration(this.typeChecker, n).kind],
                "|", ts.SyntaxKind[n.kind],
                "|", JSON.stringify(this.getCType(n))
            ));
        */
    };
    /** Mostly used inside inferTypes */
    TypeHelper.prototype.registerSyntheticNode = function (n, t) {
        if (!n || !(n.flags & ts.NodeFlags.Synthesized))
            return false;
        n.end = TypeHelper.syntheticNodesCounter++;
        this.setNodeType(n, t);
    };
    TypeHelper.prototype.setNodeType = function (n, t) {
        if (n && t)
            this.typeOfNodeDict[n.pos + "_" + n.end] = { node: n, type: t };
    };
    TypeHelper.prototype.getDeclaration = function (n) {
        var s = this.typeChecker.getSymbolAtLocation(n);
        return s && s.valueDeclaration;
    };
    TypeHelper.prototype.ensureNoTypeDuplicates = function (t) {
        if (!t)
            return null;
        var typeBodyText = getTypeBodyText(t);
        var type = this.typesDict[typeBodyText];
        if (type instanceof ArrayType)
            type.capacity = Math.max(type.capacity, t.capacity);
        if (!type)
            type = this.typesDict[typeBodyText] = t;
        return type;
    };
    /** Convert ts.Type to CType */
    TypeHelper.prototype.convertType = function (tsType, node) {
        if (!tsType || tsType.flags == ts.TypeFlags.Void)
            return exports.VoidType;
        if (tsType.flags == ts.TypeFlags.String || tsType.flags == ts.TypeFlags.StringLiteral)
            return exports.StringVarType;
        if (tsType.flags == ts.TypeFlags.Number || tsType.flags == ts.TypeFlags.NumberLiteral)
            return exports.NumberVarType;
        if (tsType.flags == ts.TypeFlags.Boolean || tsType.flags == (ts.TypeFlags.Boolean + ts.TypeFlags.Union))
            return exports.BooleanVarType;
        if (tsType.flags & ts.TypeFlags.Object && tsType.getProperties().length > 0) {
            var structType = this.generateStructure(tsType);
            var baseType = this.typeChecker.getBaseTypeOfLiteralType(tsType);
            var cTypeTag = baseType && baseType.symbol && baseType.symbol.getJsDocTags().filter(function (t) { return t.name == "ctype"; })[0];
            structType.forcedType = cTypeTag && cTypeTag.text.trim();
            structType.external = baseType && baseType.symbol && findParentSourceFile(baseType.symbol.declarations[0]).isDeclarationFile;
            return structType;
        }
        return null;
    };
    TypeHelper.prototype.generateStructure = function (tsType) {
        var userStructInfo = {};
        for (var _i = 0, _a = tsType.getProperties(); _i < _a.length; _i++) {
            var prop = _a[_i];
            if (prop.name == "prototype")
                continue;
            var declaration = prop.valueDeclaration;
            var propTsType = this.typeChecker.getTypeOfSymbolAtLocation(prop, declaration);
            var propType = this.convertType(propTsType, declaration.name) || exports.PointerVarType;
            if (propType == exports.PointerVarType && ts.isPropertyAssignment(declaration)) {
                if (declaration.initializer && ts.isArrayLiteralExpression(declaration.initializer))
                    propType = this.determineArrayType(declaration.initializer);
            }
            userStructInfo[prop.name] = { type: propType, order: declaration.pos };
        }
        return this.ensureNoTypeDuplicates(new StructType(userStructInfo));
    };
    TypeHelper.prototype.determineArrayType = function (arrLiteral) {
        var elementType = exports.PointerVarType;
        var cap = arrLiteral.elements.length;
        if (cap > 0)
            elementType = this.convertType(this.typeChecker.getTypeAtLocation(arrLiteral.elements[0])) || exports.PointerVarType;
        var type = new ArrayType(elementType, cap, false);
        this.arrayLiteralsTypes[arrLiteral.pos] = type;
        return type;
    };
    TypeHelper.prototype.mergeTypes = function (type1, type2) {
        var type1_result = { type: this.ensureNoTypeDuplicates(type1), replaced: true };
        var type2_result = { type: this.ensureNoTypeDuplicates(type2), replaced: true };
        var noChanges = { type: this.ensureNoTypeDuplicates(type1), replaced: false };
        if (!type1 && type2)
            return type2_result;
        else if (type1 && !type2)
            return type1_result;
        else if (!type1 && !type2)
            return noChanges;
        else if (typeof type1 == "string" && typeof type2 == "string" && type1 == type2)
            return noChanges;
        else if (type1 === exports.VoidType)
            return type2_result;
        else if (type2 === exports.VoidType)
            return type1_result;
        else if (type1 === exports.PointerVarType)
            return type2_result;
        else if (type2 === exports.PointerVarType)
            return type1_result;
        else if (type1 === exports.UniversalVarType)
            return type1_result;
        else if (type2 === exports.UniversalVarType)
            return type2_result;
        else if (type1 === exports.StringVarType && type2 instanceof StructType) {
            if (Object.keys(type2.properties).length == 1 && (type2.properties["length"] == exports.PointerVarType || type2.properties["length"] == exports.NumberVarType))
                return type1_result;
        }
        else if (type1 instanceof StructType && type2 === exports.StringVarType) {
            if (Object.keys(type1.properties).length == 1 && (type1.properties["length"] == exports.PointerVarType || type1.properties["length"] == exports.NumberVarType))
                return type2_result;
        }
        else if (type1 instanceof ArrayType && type2 instanceof ArrayType) {
            var cap = Math.max(type2.capacity, type1.capacity);
            var isDynamicArray = type2.isDynamicArray || type1.isDynamicArray;
            var elementTypeMergeResult = this.mergeTypes(type1.elementType, type2.elementType);
            if (type1.capacity != cap || type2.capacity != cap
                || type1.isDynamicArray != isDynamicArray || type2.isDynamicArray != isDynamicArray
                || elementTypeMergeResult.replaced)
                return { type: this.ensureNoTypeDuplicates(new ArrayType(elementTypeMergeResult.type, cap, isDynamicArray)), replaced: true };
            return noChanges;
        }
        else if (type1 instanceof DictType && type2 instanceof ArrayType) {
            return type1_result;
        }
        else if (type1 instanceof ArrayType && type2 instanceof DictType) {
            return type2_result;
        }
        else if (type1 instanceof StructType && type2 instanceof StructType) {
            var props = Object.keys(type1.properties).concat(Object.keys(type2.properties));
            var changed = false;
            var newProps = {};
            for (var _i = 0, props_1 = props; _i < props_1.length; _i++) {
                var p = props_1[_i];
                var result = this.mergeTypes(type1.properties[p], type2.properties[p]);
                var order = Math.max(type1.propertyDefs[p] ? type1.propertyDefs[p].order : 0, type2.propertyDefs[p] ? type2.propertyDefs[p].order : 0);
                newProps[p] = { type: result.type, order: order };
                if (result.replaced)
                    changed = true;
            }
            return changed ? { type: this.ensureNoTypeDuplicates(new StructType(newProps)), replaced: true } : noChanges;
        }
        else if (type1 instanceof ArrayType && type2 instanceof StructType) {
            return this.mergeArrayAndStruct(type1, type2);
        }
        else if (type1 instanceof StructType && type2 instanceof ArrayType) {
            return this.mergeArrayAndStruct(type2, type1);
        }
        else if (type1 instanceof DictType && type2 instanceof StructType) {
            return this.mergeDictAndStruct(type1, type2);
        }
        else if (type1 instanceof StructType && type2 instanceof DictType) {
            return this.mergeDictAndStruct(type2, type1);
        }
        else if (type1 instanceof DictType && type2 instanceof DictType) {
            var _a = this.mergeTypes(type1.elementType, type2.elementType), elemType = _a.type, replaced = _a.replaced;
            if (replaced)
                return { type: this.ensureNoTypeDuplicates(new DictType(elemType)), replaced: true };
            else
                return noChanges;
        }
        else if (type1 instanceof FuncType && type2 instanceof FuncType) {
            var _b = this.mergeTypes(type1.returnType, type2.returnType), returnType = _b.type, returnTypeReplaced = _b.replaced;
            var _c = this.mergeTypes(type1.instanceType, type2.instanceType), instanceType = _c.type, instanceTypeReplaced = _c.replaced;
            var paramCount = Math.max(type1.parameterTypes.length, type2.parameterTypes.length);
            var paramTypesReplaced = type1.parameterTypes.length !== type2.parameterTypes.length;
            var paramTypes = [];
            for (var i_5 = 0; i_5 < paramCount; i_5++) {
                var _d = this.mergeTypes(type1.parameterTypes[i_5], type2.parameterTypes[i_5]), pType = _d.type, pTypeReplaced = _d.replaced;
                paramTypes.push(pType);
                if (pTypeReplaced)
                    paramTypesReplaced = true;
            }
            if (returnTypeReplaced || instanceTypeReplaced || paramTypesReplaced)
                return { type: this.ensureNoTypeDuplicates(new FuncType(returnType, paramTypes, instanceType)), replaced: true };
            else
                return noChanges;
        }
        else
            return { type: exports.UniversalVarType, replaced: true };
    };
    TypeHelper.prototype.mergeArrayAndStruct = function (arrayType, structType) {
        var props = Object.keys(structType.properties);
        var needPromoteToDictionary = false;
        var needPromoteToTuple = false;
        for (var _i = 0, props_2 = props; _i < props_2.length; _i++) {
            var p = props_2[_i];
            if (p == "length")
                continue;
            if (isNaN(+p))
                needPromoteToDictionary = true;
            if (this.mergeTypes(arrayType.elementType, structType.properties[p]).replaced)
                needPromoteToTuple = true;
        }
        if (needPromoteToDictionary && needPromoteToTuple)
            return { type: this.ensureNoTypeDuplicates(new DictType(exports.UniversalVarType)), replaced: true };
        else if (needPromoteToDictionary)
            return { type: this.ensureNoTypeDuplicates(new DictType(arrayType.elementType)), replaced: true };
        else if (needPromoteToTuple)
            return { type: this.ensureNoTypeDuplicates(new ArrayType(exports.UniversalVarType, arrayType.capacity, arrayType.isDynamicArray)), replaced: true };
        else
            return { type: arrayType, replaced: true };
    };
    TypeHelper.prototype.mergeDictAndStruct = function (dictType, structType) {
        var elementType = dictType.elementType;
        for (var k in structType.properties)
            (elementType = this.mergeTypes(elementType, structType.properties[k]).type);
        return { type: this.ensureNoTypeDuplicates(new DictType(elementType)), replaced: true };
    };
    TypeHelper.syntheticNodesCounter = 0;
    return TypeHelper;
}());
exports.TypeHelper = TypeHelper;
//# sourceMappingURL=types.js.map